from pyetg1510.mailbox.connection import *
from pyetg1510.mailbox.mailbox_gateway import *
from pyetg1510.mailbox.sdo_application_interface import *

VERSION = (0, 0, 1)

__version__ = ".".join([str(x) for x in VERSION])
